import anndata 
import torch
import scanpy as sc
import pandas as pd
import numpy as np
from .exp import VaeSmExperiment
from plotnine import *
import plotly.graph_objects as go
import plotly.io as pio
from .commons import make_edge_df
from matplotlib import colors

def safe_toarray(x):
    if type(x) != np.ndarray:
        return x.toarray()
    else:
        return x

def make_inputs(sc_adata, sp_adata, layer_name):
    x = torch.tensor(safe_toarray(sc_adata.layers[layer_name]))
    s = torch.tensor(safe_toarray(sp_adata.layers[layer_name]))
    if (x - x.int()).norm() > 0:
        try:
            raise ValueError('target layer of sc_adata should be raw count')
        except ValueError as e:
           print(e) 
    if (s - s.int()).norm() > 0:
        try:
            raise ValueError('target layer of sp_adata should be raw count')
        except ValueError as e:
           print(e) 
    return x, s

def optimize_deepcolor(vaesm_exp, lr, x_batch_size, s_batch_size, first_epoch, second_epoch):
    print(f'Loss: {vaesm_exp.evaluate()}')
    print('Start first opt')
    vaesm_exp.mode_change('sc')
    vaesm_exp.initialize_optimizer(lr)
    vaesm_exp.initialize_loader(x_batch_size, s_batch_size)
    vaesm_exp.train_total(first_epoch)
    print('Done first opt')
    print(f'Loss: {vaesm_exp.evaluate()}')
    print('Start second opt')
    vaesm_exp.mode_change('sp')
    vaesm_exp.initialize_optimizer(lr)
    vaesm_exp.initialize_loader(x_batch_size, s_batch_size)
    vaesm_exp.train_total(second_epoch)
    print('Done second opt')
    print(f'Loss: {vaesm_exp.evaluate()}')
    return vaesm_exp


def conduct_umap(adata, key):
    sc.pp.neighbors(adata, use_rep=key, n_neighbors=30)
    sc.tl.umap(adata)
    return(adata)


def extract_mapping_info(vaesm_exp, sc_adata, sp_adata):
    with torch.no_grad():
        xz, qxz, xld, p, sld, theta_x, theta_s = vaesm_exp.vaesm(vaesm_exp.xedm.x.to(vaesm_exp.device))
    sc_adata.obsm['X_zl'] = qxz.loc.detach().cpu().numpy()
    sc_adata.obsm['lambda'] = xld.detach().cpu().numpy()
    p_df = pd.DataFrame(p.detach().cpu().numpy().transpose(), index=sc_adata.obs_names, columns=sp_adata.obs_names)
    sc_adata.obsm['map2sp'] = p_df.values
    sp_adata.obsm['map2sc'] = p_df.transpose().values
    sc_adata.obsm['p_mat'] = sc_adata.obsm['map2sp'] / np.sum(sc_adata.obsm['map2sp'], axis=1).reshape((-1, 1))
    return sc_adata, sp_adata


def estimate_spatial_distribution(
        sc_adata, sp_adata, param_save_path, layer_name='count', first_epoch=500, second_epoch=500, lr=0.001, val_ratio=0.01, test_ratio=0.01, device=None, num_workers=1,
        x_batch_size=1000, s_batch_size=100, 
        model_params = {
            "x_dim": 100,
            "s_dim": 100,
            "xz_dim": 10, "sz_dim": 10,
            "enc_z_h_dim": 50, "dec_z_h_dim": 50, "map_h_dim": 50,
            "num_enc_z_layers": 2, "num_dec_z_layers": 2
        }
    ):
    if device == None:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
    # make data set 
    sp_adata.obs_names_make_unique()
    sc_adata.obs_names_make_unique()
    x, s = make_inputs(sc_adata, sp_adata, layer_name)
    model_params['x_dim'] = x.size()[1]
    model_params['s_dim'] = s.size()[1]
    vaesm_exp = VaeSmExperiment(model_params, lr, x, s, test_ratio, 100, 100, num_workers, validation_ratio=val_ratio, device=device)
    vaesm_exp = optimize_deepcolor(vaesm_exp, lr, x_batch_size, s_batch_size, first_epoch, second_epoch)
    torch.save(vaesm_exp.vaesm.state_dict(), param_save_path)
    sc_adata.uns['param_save_path'] = param_save_path
    sp_adata.uns['param_save_path'] = param_save_path
    sp_adata.uns['layer_name'] = layer_name
    sc_adata, sp_adata = extract_mapping_info(vaesm_exp, sc_adata, sp_adata)
    return sc_adata, sp_adata


def calculate_clusterwise_distribution(sc_adata, sp_adata, cluster_label):
    p_mat = sp_adata.obsm['map2sc']
    celltypes = np.sort(np.unique(sc_adata.obs[cluster_label]))
    try:
        raise ValueError('some of cluster names in `cluster_label` is overlapped with `sp_adata.obs.columns`')
    except ValueError as e:
        print(e)
    cp_map_df = pd.DataFrame({
        celltype: np.sum(p_mat[:, sc_adata.obs[cluster_label] == celltype], axis=1)
        for celltype in celltypes}, index=sp_adata.obs_names)
    sp_adata.obs = pd.concat([sp_adata.obs, cp_map_df], axis=1)
    return sp_adata

def calculate_imputed_spatial_expression(sc_adata, sp_adata):
    sc_norm_mat = sc_adata.layers['count'].toarray() / np.sum(sc_adata.layers['count'].toarray(), axis=1).reshape((-1, 1))
    sp_adata.layers['imputed_exp'] = np.matmul(
        sp_adata.obsm['map2sc'], sc_norm_mat)
    return sp_adata

def estimate_colocalization(sc_adata):
    p_mat = sc_adata.obsm['p_mat']
    coloc_mat = p_mat @ p_mat.transpose()
    coloc_mat = np.log2(coloc_mat) + np.log2(p_mat.shape[1])
    sc_adata.obsp['colocalization'] = coloc_mat
    return sc_adata

def make_coloc_mat(sc_adata):
    p_mat = sc_adata.obsm['p_mat']
    coloc_mat = p_mat @ p_mat.transpose()
    coloc_mat = np.log2(coloc_mat) + np.log2(p_mat.shape[1])
    return coloc_mat


def make_high_coloc_index(sc_adata, celltype_label):
    coloc_mat = make_coloc_mat(sc_adata) 
    thresh = 1
    high_coloc_index = np.argwhere(coloc_mat > thresh)
    high_coloc_index = high_coloc_index[high_coloc_index[:, 0] < high_coloc_index[:, 1]]
    ocell1_types = sc_adata.obs[celltype_label].iloc[high_coloc_index[:, 0]].values
    ocell2_types = sc_adata.obs[celltype_label].iloc[high_coloc_index[:, 1]].values
    high_coloc_index = high_coloc_index[ocell1_types != ocell2_types]
    return high_coloc_index

def make_cell_umap_df(cell, edge_df, sc_adata):
    cell_adata = sc_adata[edge_df[cell]]
    cell_umap_df = pd.DataFrame(cell_adata.obsm['position'] * 0.9, columns=['X', 'Y'], index=edge_df.index)
    cell_umap_df['edge'] = edge_df.index
    return(cell_umap_df)

def make_edge_vis_df(sc_adata, celltype_label, total_edge_num, edge_thresh=1):
    orig_edge_df = make_edge_df(sc_adata, celltype_label, edge_thresh=edge_thresh)
    sub_edge_df = orig_edge_df.loc[np.random.choice(orig_edge_df.index, total_edge_num)]
    tot_edge_df = pd.concat([
        make_cell_umap_df(cell, sub_edge_df, sc_adata)
        for cell in ['cell1', 'cell2']], axis=0)
    return tot_edge_df

def visualize_colocalization_network(sc_adata, sp_adata, celltype_label, spatial_cluster, celltype_sample_num=500, total_edge_num=5000,  color_dict=None, edge_thresh=1):
    # resample celltypes
    sc_adata = sc_adata[sc_adata.obs.groupby(celltype_label).sample(celltype_sample_num, replace=True).index]
    sc_adata.obs_names_make_unique()
    # determine cell thetas and positions
    thetas = 2 * np.pi * np.arange(sc_adata.shape[0]) / sc_adata.shape[0]
    x = np.cos(thetas)
    y = np.sin(thetas)
    pos_mat = np.column_stack((x, y))
    sc_adata.obsm['position'] = pos_mat
    # map max leiden
    p_mat = sc_adata.obsm['p_mat']
    sc_adata.obs['max_map'] = sp_adata[p_mat.argmax(axis=1)].obs[spatial_cluster].astype(str).values
    # sample cell pairs based on colocalization
    tot_edge_df = make_edge_vis_df(sc_adata, celltype_label, total_edge_num, edge_thresh=edge_thresh)
    # visualize
    cells_df = pd.DataFrame({
        'X': sc_adata.obsm['position'][:, 0] * np.random.uniform(0.9, 1.1, size=sc_adata.shape[0]), 
        'Y': sc_adata.obsm['position'][:, 1] * np.random.uniform(0.9, 1.1, size=sc_adata.shape[0]), 
        'celltype': sc_adata.obs[celltype_label]})
    # determine even odds groups
    groups = sc_adata.obs[celltype_label].unique()
    gidxs = np.arange(groups.shape[0])
    even_groups = groups[gidxs % 2 == 0]
    odd_groups = groups[gidxs % 2 == 1]
    even_cells_df = cells_df.query('celltype in @even_groups')
    odd_cells_df = cells_df.query('celltype in @odd_groups')
    celltype_df = cells_df.groupby('celltype', as_index=False).mean()
    add_df = pd.DataFrame({
        'X': sc_adata.obsm['position'][:, 0] * np.random.uniform(1.1, 1.3, size=sc_adata.shape[0]), 
        'Y': sc_adata.obsm['position'][:, 1] * np.random.uniform(1.1, 1.3, size=sc_adata.shape[0]),
        'celltype': sc_adata.obs['max_map'].astype(str)})
    g = ggplot(add_df, aes(x='X', y='Y', color='celltype')) + geom_point(size=0.5)  +\
         geom_point(even_cells_df, size=0.1, color='#60C2CB') + \
          geom_point(odd_cells_df, size=0.1, color='#D05C54')  + \
          geom_line(tot_edge_df, aes(group='edge'), color='black', size=0.1, alpha=0.05) + \
              geom_text(celltype_df, aes(label='celltype'), color='black')  
    if not color_dict == None:
        g = g + scale_color_manual(color_dict)
    return g


# spcify top expression
def make_top_values(mat, top_fraction = 0.1, axis=0):
    top_mat = mat > np.quantile(mat, 1 - top_fraction, axis=axis, keepdims=True)
    return(top_mat)


def make_top_act_ligands(cell_type, coexp_count_df, topn=3):
    d = coexp_count_df.loc[cell_type].max(axis=0).sort_values(ascending=False)[:topn]
    return(d.index)


def make_coexp_cc_df(ligand_adata, edge_df, role):
    sender = edge_df.cell1 if role == "sender" else edge_df.cell2
    receiver = edge_df.cell2 if role == "sender" else edge_df.cell1
    coexp_df = pd.DataFrame(
        ligand_adata[sender].X *
        ligand_adata[receiver].layers['activity'],
        columns=ligand_adata.var_names, index=edge_df.index
    )
    coexp_df['cell2_type'] = edge_df['cell2_type']
    coexp_df['cell1_type'] = edge_df['cell1_type']
    coexp_cc_df = coexp_df.groupby(['cell2_type', 'cell1_type']).sum()
    coexp_cc_df = coexp_cc_df.reset_index().melt(id_vars=['cell1_type', 'cell2_type'], var_name='ligand', value_name='coactivity')
    return coexp_cc_df

def calculate_proximal_cell_communications(sc_adata, celltype_label, lt_df, target_cellset, celltype_sample_num=500, ntop_genes=4000, each_display_num=3, role="sender", edge_thresh=1, target_color_dict=None):
    # subsample data
    sc_adata = sc_adata[sc_adata.obs.groupby(celltype_label).sample(celltype_sample_num, replace=True).index]
    sc_adata.obs_names_make_unique()
    celltypes = sc_adata.obs.loc[:, celltype_label].unique()
    # make edge_df
    edge_df = make_edge_df(sc_adata, celltype_label, sub_sample=False, exclude_reverse=False, edge_thresh=edge_thresh)
    # select edge df with cell1 as target
    edge_df = edge_df.loc[edge_df.cell1_type.isin(target_cellset)]
    edge_df = edge_df.loc[~edge_df.cell2_type.isin(target_cellset)]
    # select genes
    sc.pp.highly_variable_genes(sc_adata, n_top_genes=ntop_genes)
    sc_adata = sc_adata[:, sc_adata.var.highly_variable]
    common_genes = np.intersect1d(lt_df.index, sc_adata.var_names)
    lt_df = lt_df.loc[common_genes]
    sc_adata = sc_adata[:, common_genes]
    lt_df = lt_df.loc[:, lt_df.columns.isin(sc_adata.var_names)]
    # make normalization
    ligands = lt_df.columns
    lt_df = lt_df.div(lt_df.sum(axis=0), axis=1)
    ligand_adata = sc_adata[:, ligands]
    top_exps = make_top_values(sc_adata.X.toarray(), axis=1, top_fraction=0.01)
    ligand_adata.layers['activity'] = make_top_values(top_exps @ lt_df)
    ligand_adata.X = make_top_values(safe_toarray(ligand_adata.X))
    # make base coexp df
    coexp_cc_df = make_coexp_cc_df(ligand_adata, edge_df, role)
    sub_coexp_cc_df = coexp_cc_df.sort_values('coactivity', ascending=False).groupby('cell2_type', as_index=False).head(n=each_display_num)
    # plotting configurrations
    tot_list = list(sub_coexp_cc_df.ligand.unique()) + list(celltypes)
    ligand_pos_dict = pd.Series({
        ligand: i
        for i, ligand in enumerate(sub_coexp_cc_df.ligand.unique())
    })
    celltype_pos_dict = pd.Series({
        celltype: i + sub_coexp_cc_df.ligand.unique().shape[0]
        for i, celltype in enumerate(celltypes)
    })
    senders = sub_coexp_cc_df.cell1_type.values if role == "sender" else sub_coexp_cc_df.cell2_type.values
    receivers = sub_coexp_cc_df.cell2_type.values if role == "sender" else sub_coexp_cc_df.cell1_type.values
    sources = pd.concat([ligand_pos_dict.loc[sub_coexp_cc_df.ligand.values], celltype_pos_dict.loc[senders]])
    targets = pd.concat([celltype_pos_dict.loc[receivers], ligand_pos_dict.loc[sub_coexp_cc_df.ligand.values]])
    values = pd.concat([sub_coexp_cc_df['coactivity'], sub_coexp_cc_df['coactivity']])
    labels = pd.concat([sub_coexp_cc_df['cell1_type'], sub_coexp_cc_df['cell1_type']])
    # colors = pd.Series(target_color_dict)[labels]
    fig = go.Figure(data=[go.Sankey(node=dict(label=tot_list),
        link=dict(
            source=sources,
            target=targets,
            value=values,
            # color=colors,
            label=labels))])
    fig.update_layout(font_family="Courier New")
    return fig, coexp_cc_df


def make_dual_zl(sc_adata, high_coloc_index):
    dual_zl_add = sc_adata.obsm['X_zl'][high_coloc_index[:, 0]] + sc_adata.obsm['X_zl'][high_coloc_index[:, 1]]
    dual_zl_prod = sc_adata.obsm['X_zl'][high_coloc_index[:, 0]] * sc_adata.obsm['X_zl'][high_coloc_index[:, 1]]
    dual_zl_prod = np.sign(dual_zl_prod) * np.sqrt(np.abs(dual_zl_prod))
    dual_zl = np.concatenate([dual_zl_add, dual_zl_prod], axis=1)
    return dual_zl_add


def setup_dual_adata(dual_zl, sc_adata, high_coloc_index):
    dual_adata = anndata.AnnData(dual_zl)
    dual_adata.obsm['X_zl'] = dual_zl
    dual_adata.obs['cell1_celltype'] = sc_adata.obs["large_class"].values[high_coloc_index[:, 0]]
    dual_adata.obs['cell2_celltype'] = sc_adata.obs["large_class"].values[high_coloc_index[:, 1]]
    dual_adata.obs['cell1_obsname'] = sc_adata.obs_names[high_coloc_index[:, 0]]
    dual_adata.obs['cell2_obsname'] = sc_adata.obs_names[high_coloc_index[:, 1]]
    dual_adata.obs_names = dual_adata.obs['cell1_obsname'] + dual_adata.obs['cell2_obsname']
    cell_min = dual_adata.obs[['cell1_celltype', 'cell2_celltype']].astype(str).values.min(axis=1)
    cell_max = dual_adata.obs[['cell1_celltype', 'cell2_celltype']].astype(str).values.max(axis=1)
    dual_adata.obs['dual_celltype'] = cell_min + '/' + cell_max
    return dual_adata


def analyze_pair_cluster(sc_adata, sp_adata, cellset1, cellset2, celltype_label, max_pair_num=30000):
    contributions = np.sum(sc_adata.obsm['map2sp'], axis=1)
    sc_adata.obs['large_class'] = 'None'
    annot1 = ','.join(cellset1)
    annot2 = ','.join(cellset2)
    sc_adata.obs['large_class'][sc_adata.obs[celltype_label].isin(cellset1)] = annot1
    sc_adata.obs['large_class'][sc_adata.obs[celltype_label].isin(cellset2)] = annot2
    sc_adata = sc_adata[sc_adata.obs['large_class'].isin([annot1, annot2])]
    high_coloc_index = make_high_coloc_index(sc_adata, "large_class")
    if high_coloc_index.shape[0] > max_pair_num:
        high_coloc_index = high_coloc_index[np.random.randint(0, high_coloc_index.shape[0], size=max_pair_num)]
    dual_zl = make_dual_zl(sc_adata, high_coloc_index)
    dual_adata = setup_dual_adata(dual_zl, sc_adata, high_coloc_index)
    dual_adata = conduct_umap(dual_adata, 'X_zl')
    sc.tl.leiden(dual_adata, resolution=0.1)
    dual_adata.uns['large_class1'] = annot1
    dual_adata.uns['large_class2'] = annot2
    return dual_adata



# evndyn
import torch
from .modules import VaeSm, scVAE
from .funcs import calc_kld, calc_nb_loss
from .dataset import VaeSmDataSet, VaeSmDataManager, VaeSmDataManagerDPP, ConcatDataset
from torch.utils.data import DataLoader
import numpy as np
import copy
import torch.nn as nn


class VaeSmExperiment:
    def __init__(self, model_params, lr, x, s, test_ratio, x_batch_size, s_batch_size, num_workers, validation_ratio=0.1, device='auto'):
        if device == 'auto':
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = torch.device(device)
        self.sedm = VaeSmDataManager(s, test_ratio, x_batch_size, num_workers, validation_ratio=validation_ratio)
        self.xedm = VaeSmDataManager(x, test_ratio, s_batch_size, num_workers, validation_ratio=validation_ratio)
        self.model_params = model_params
        self.vaesm = VaeSm(self.sedm.x.size()[0], **self.model_params)
        self.vaesm.to(self.device)
        self.vaesm_optimizer = torch.optim.Adam(self.vaesm.parameters(), lr=lr)
        self.train_loss_list = []
        self.test_loss_list = []
        self.best_loss = None
        s = self.sedm.x
        snorm_mat = self.sedm.xnorm_mat
        self.s = s.to(self.device)
        self.snorm_mat = snorm_mat.to(self.device)
        self.mode = 'sc'

    def elbo_loss(self, x, xnorm_mat, s, snorm_mat):
        xz, qxz, xld, p, sld, theta_x, theta_s = self.vaesm(x)
        elbo_loss = 0
        if self.mode != 'sp':        
            # kld of pz and qz
            elbo_loss += calc_kld(qxz).sum()
            # reconst loss of x
            elbo_loss += calc_nb_loss(xld, xnorm_mat, theta_x, x).sum()
        if self.mode != 'sc':            
            # reconst loss of s
            elbo_loss += calc_nb_loss(sld, snorm_mat, theta_s, s).sum()
        return(elbo_loss)
        
    def train_epoch(self):
        total_loss = 0
        entry_num = 0
        for x, xnorm_mat in self.xedm.train_loader:
            x = x.to(self.device)
            xnorm_mat = xnorm_mat.to(self.device)
            self.vaesm_optimizer.zero_grad()
            loss = self.elbo_loss(
                x, xnorm_mat, self.s, self.snorm_mat)
            loss.backward()
            self.vaesm_optimizer.step()
        return(0)

    def evaluate(self, mode = 'test'):
        with torch.no_grad():
            self.vaesm.eval()
            if mode == 'test':            
                x = self.xedm.test_x.to(self.device)
                xnorm_mat = self.xedm.test_xnorm_mat.to(self.device)
            else:
                x = self.xedm.validation_x.to(self.device)
                xnorm_mat = self.xedm.validation_xnorm_mat.to(self.device)            
            loss = self.elbo_loss(
                x, xnorm_mat, self.s, self.snorm_mat)
            entry_num = x.shape[0]
            loss_val = loss / entry_num
        return(loss_val)

    def train_total(self, epoch_num):
        self.vaesm.train()
        for epoch in range(epoch_num):
            state_dict = copy.deepcopy(self.vaesm.state_dict())
            loss = self.train_epoch()
            if np.isnan(loss):
                self.vaesm.load_state_dict(state_dict)
                break
            if epoch % 10 == 0:
                loss = self.evaluate(mode='validation')
                print(f'loss at epoch {epoch} is {loss}')

    def initialize_optimizer(self, lr):
        self.vaesm_optimizer = torch.optim.Adam(self.vaesm.parameters(), lr=lr)

    def initialize_loader(self, x_batch_size, s_batch_size):
        self.xedm.initialize_loader(x_batch_size)
        self.sedm.initialize_loader(s_batch_size)

    def mode_change(self, mode):
        self.mode = mode
        if mode == 'sc':
            self.vaesm.sc_mode()        
        if mode == 'sp':
            self.vaesm.sp_mode()        
        if mode == 'dual':
            self.vaesm.dual_mode()
        
class VaeSmExperimentMB(VaeSmExperiment):
    def __init__(self, model_params, lr, x, s, test_ratio, x_batch_size, s_batch_size, num_workers, validation_ratio=0.1, device='auto'):
        super(VaeSmExperimentMB, self).__init__()

    def train_epoch(self):
        total_loss = 0
        entry_num = 0
        for (x, xnorm_mat, x_batch_idx), (s, snorm_mat, s_batch_idx) in zip(self.xedm.train_loader, self.sedm.train_loader):
            x = x.to(self.device)
            xnorm_mat = xnorm_mat.to(self.device)
            x_batch_idx = x_batch_idx.to(self.device)
            s = s.to(self.device)
            s_batch_idx = s_batch_idx.to(self.device)
            self.vaesm_optimizer.zero_grad()
            loss = self.elbo_loss(
                x, xnorm_mat, s, snorm_mat)
            loss.backward()
            self.vaesm_optimizer.step()
        return(0)

    def evaluate(self, mode = 'test'):
        with torch.no_grad():
            self.vaesm.eval()
            if mode == 'test':            
                x = self.xedm.test_x.to(self.device)
                xnorm_mat = self.xedm.test_xnorm_mat.to(self.device)
                s = self.sedm.test_x.to(self.device)
                snorm_mat = self.sedm.test_xnorm_mat.to(self.device)
            else:
                x = self.xedm.validation_x.to(self.device)
                xnorm_mat = self.xedm.validation_xnorm_mat.to(self.device)            
                s = self.sedm.validation_x.to(self.device)
                snorm_mat = self.sedm.validation_xnorm_mat.to(self.device)
            loss = self.elbo_loss(
                x, xnorm_mat, s, snorm_mat)
            entry_num = x.shape[0]
            loss_val = loss / entry_num
        return(loss_val)

class VaeSmExperimentDPP(VaeSmExperiment):
    def __init__(self, gpu, gpu_num, model_params, lr, x, s, test_ratio, x_batch_size, s_batch_size, num_workers, validation_ratio=0.1, device=None):
        # Wrap the model
        self.device = torch.device(f'cuda:{gpu}')
        self.gpu = gpu
        torch.cuda.set_device(gpu)
        self.model_params = model_params
        self.vaesm = VaeSm(s.size()[0],  **self.model_params)
        self.vaesm.to(self.device)
        self.vaesm = nn.parallel.DistributedDataParallel(self.vaesm, device_ids=[gpu], find_unused_parameters=True)
        self.sedm = VaeSmDataManagerDPP(gpu, gpu_num, s, test_ratio, x_batch_size, num_workers, validation_ratio=validation_ratio)
        self.xedm = VaeSmDataManagerDPP(gpu, gpu_num, x, test_ratio, s_batch_size, num_workers, validation_ratio=validation_ratio)
        self.vaesm_optimizer = torch.optim.Adam(self.vaesm.parameters(), lr=lr)
        self.train_loss_list = []
        self.test_loss_list = []
        self.best_loss = None
        s = self.sedm.x
        snorm_mat = self.sedm.xnorm_mat
        self.s = s.to(self.device)
        self.snorm_mat = snorm_mat.to(self.device)
        self.mode = 'sc'

    def initialize_optimizer(self, lr):
        self.vaesm_optimizer = torch.optim.Adam(self.vaesm.parameters(), lr=lr)

    def initialize_loader(self, x_batch_size, s_batch_size):
        self.xedm.initialize_loader(x_batch_size)
        self.sedm.initialize_loader(s_batch_size)

    def mode_change(self, mode):
        state_dict = copy.deepcopy(self.vaesm.module.state_dict())
        self.vaesm = VaeSm(self.sedm.x.size()[0], **self.model_params)
        self.vaesm.load_state_dict(state_dict)
        self.vaesm.to(self.device)
        self.mode = mode
        if mode == 'sc':
            self.vaesm.sc_mode()        
        if mode == 'sp':
            self.vaesm.sp_mode()        
        if mode == 'dual':
            self.vaesm.dual_mode()
        self.vaesm = nn.parallel.DistributedDataParallel(self.vaesm, device_ids=[self.gpu], find_unused_parameters=True)
        
    



import torch
import torch.distributions as dist
import torch.nn as nn
from torch.nn.parameter import Parameter
from torch import functional as F
from torch.distributions.kl import kl_divergence
from torch.nn import init
import numpy as np

class LinearReLU(nn.Module):
    def __init__(self, input_dim, output_dim):
        super(LinearReLU, self).__init__()
        self.f = nn.Sequential(
            nn.Linear(input_dim, output_dim),
            nn.LayerNorm(output_dim, elementwise_affine=False),
            nn.ReLU(True))

    def forward(self, x):
        h = self.f(x)
        return(h)


class SeqNN(nn.Module):
    def __init__(self, num_steps, dim):
        super(SeqNN, self).__init__()
        modules = [
            LinearReLU(dim, dim)
            for _ in range(num_steps)
        ]
        self.f = nn.Sequential(*modules)

    def forward(self, pre_h):
        post_h = self.f(pre_h)
        return(post_h)


class Encoder(nn.Module):
    def __init__(self, num_h_layers, x_dim, h_dim, z_dim):
        super(Encoder, self).__init__()
        self.x2h = LinearReLU(x_dim, h_dim)
        self.seq_nn = SeqNN(num_h_layers - 1, h_dim)
        self.h2mu = nn.Linear(h_dim, z_dim)
        self.h2logvar = nn.Linear(h_dim, z_dim)

    def forward(self, x):
        pre_h = self.x2h(x)
        post_h = self.seq_nn(pre_h)
        mu = self.h2mu(post_h)
        logvar = self.h2logvar(post_h)
        return(mu, logvar)


class Decoder(nn.Module):
    def __init__(self, num_h_layers, z_dim, h_dim, x_dim):
        super(Decoder, self).__init__()
        self.z2h = LinearReLU(z_dim, h_dim)
        self.seq_nn = SeqNN(num_h_layers - 1, h_dim)
        self.h2ld = nn.Linear(h_dim, x_dim)
        self.softplus = nn.Softplus()

    def forward(self, z):
        pre_h = self.z2h(z)
        post_h = self.seq_nn(pre_h)
        ld = self.h2ld(post_h)
        correct_ld = self.softplus(ld)
        return(correct_ld)

    
class DecoderSoftMax(nn.Module):
    def __init__(self, num_h_layers, z_dim, h_dim, x_dim):
        super(Decoder, self).__init__()
        self.z2h = LinearReLU(z_dim, h_dim)
        self.seq_nn = SeqNN(num_h_layers - 1, h_dim)
        self.h2ld = nn.Linear(h_dim, x_dim)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, z):
        pre_h = self.z2h(z)
        post_h = self.seq_nn(pre_h)
        ld = self.h2ld(post_h)
        correct_ld = self.softmax(ld)
        return(correct_ld)

    
class scVAE(nn.Module):
    def __init__(
            self,
            x_dim, xz_dim,
            enc_z_h_dim, dec_z_h_dim,
            num_enc_z_layers,
            num_dec_z_layers, **kwargs):
        super(scVAE, self).__init__()
        self.enc_z = Encoder(num_enc_z_layers, x_dim, enc_z_h_dim, xz_dim)
        self.dec_z2x = Decoder(num_enc_z_layers, xz_dim, dec_z_h_dim, x_dim)
        self.softplus = nn.Softplus()

    def forward(self, x):
        # encode z
        qz_mu, qz_logvar = self.enc_z(x)
        qz = dist.Normal(qz_mu, self.softplus(qz_logvar))
        z = qz.rsample()
        # decode z
        xld = self.dec_z2x(z)
        return(z, qz, xld)

class scVAEScale(scVAE):
    def __init__(
            self,
            x_dim, xz_dim,
            enc_z_h_dim, dec_z_h_dim,
            num_enc_z_layers,
            num_dec_z_layers, **kwargs):
        super(scVAEScale, self).__init__()
        self.dec_z2x = DecoderSoftMax(num_enc_z_layers, xz_dim, dec_z_h_dim, x_dim)
        self.logxcoeff = Parameter(torcch.Tensor(x_dim))

    def reset_parameters(self):
        init.normal_(self.logxcoeff)

    def forward(self, x):
        # encode z
        qz_mu, qz_logvar = self.enc_z(x)
        qz = dist.Normal(qz_mu, self.softplus(qz_logvar))
        z = qz.rsample()
        # decode z
        xcoeff = self.softplus(self.logxcoeff)
        xld = self.dec_z2x(z) * xcoeff
        return(z, qz, xld)

class SpatialMap(nn.Module):
    def __init__(
            self,
            sz_dim, xz_dim, h_dim):
        super(SpatialMap, self).__init__()
        self.dec_z2p = Decoder(2, sz_dim + xz_dim, h_dim, 2)
        self.softplus = nn.Softplus()
    def forward(self, sz, xz):
        ext_xz = xz.expand(sz.size()[0], *xz.size())
        ext_sz = sz.expand(xz.size()[0], *sz.size()).transpose(0, 1)
        concat_z = torch.cat([ext_sz, ext_xz], dim=2)
        attn = self.dec_z2p(concat_z)[:, :, 0]
        attn = attn / attn.sum(dim=1, keepdim=True)
        return(attn)

class VaeSm(nn.Module):
    def __init__(
            self,
            s_num, x_dim, sz_dim,  xz_dim,
            enc_z_h_dim,  dec_z_h_dim, map_h_dim,
            num_enc_z_layers, num_dec_z_layers, **kwargs):
        super(VaeSm, self).__init__()
        self.scvae = scVAE(x_dim, xz_dim, enc_z_h_dim, dec_z_h_dim, num_enc_z_layers, num_dec_z_layers)
        self.dec_xz2p = Decoder(num_enc_z_layers, xz_dim, dec_z_h_dim, s_num)
        self.logscoeff = Parameter(torch.Tensor(x_dim))
        self.logscoeff_add = Parameter(torch.Tensor(x_dim))
        self.logtheta_x =  Parameter(torch.Tensor(x_dim))
        self.logtheta_s =  Parameter(torch.Tensor(x_dim))
        self.softplus = nn.Softplus()
        self.reset_parameters()
        self.mode = 'dual'

    def reset_parameters(self):
        init.normal_(self.logscoeff)
        init.normal_(self.logscoeff_add)
        init.normal_(self.logtheta_x)
        init.normal_(self.logtheta_s)
        
    def forward(self, x):
        # encode xz
        xz, qxz, xld = self.scvae(x)
        # encode sz
        # deconst p
        p = self.dec_xz2p(xz).transpose(0, 1)
        p = p / p.sum(dim=1, keepdim=True)
        # construct mean for s
        scoeff = self.softplus(self.logscoeff)
        scoeff_add = self.softplus(self.logscoeff_add)
        sld = torch.matmul(p, xld * scoeff) + scoeff_add
        theta_x = self.softplus(self.logtheta_x)
        theta_s = self.softplus(self.logtheta_s)
        return(xz, qxz, xld, p, sld, theta_x, theta_s)

    def dual_mode(self):
        self.mode = 'dual'
        for parameter in self.parameters():
            parameter.requires_grad = True

    def sp_mode(self):
        self.mode = 'sp'
        for parameter in self.parameters():
            parameter.requires_grad = True
        for parameter in self.scvae.parameters():
            parameter.requires_grad = False
        self.logtheta_x.requires_grad = False

    def sc_mode(self):
        self.mode = 'sc'
        for parameter in self.parameters():
            parameter.requires_grad = True
        for params in [self.dec_xz2p.parameters(), [self.logscoeff, self.logscoeff_add, self.logtheta_s]]:
            for parameter in params:
                parameter.requires_grad = False
        

class VaeSmMB(VaeSm):
    def __init__(
            self, s_batch_num, x_batch_num,
            s_dim, x_dim, sz_dim,  xz_dim,
            enc_z_h_dim,  dec_z_h_dim, map_h_dim,
            num_enc_z_layers, num_dec_z_layers, **kwargs):
        super(VaeSmMB, self).__init__()
        self.s_batch_rate = Parameter(torch.Tensor(s_batch_num, s_dim))
        self.x_batch_rate = Parameter(torch.Tensor(x_batch_num, x_dim))
        self.s_batch_add = Parameter(torch.Tensor(s_batch_num, s_dim))
        self.x_batch_add = Parameter(torch.Tensor(x_batch_num, x_dim))

    def reset_parameters(self):
        init.normal_(self.logscoeff)
        init.normal_(self.s_batch_rate)
        init.normal_(self.x_batch_rate)
        init.normal_(self.s_batch_add)
        init.normal_(self.x_batch_add)
        
    def forward(self, x, x_batch_idx, s, s_batch_idx):
        # encode xz
        xz, qxz, xld = self.scvae(x)
        # encode sz
        qsz_mu, qsz_logvar = self.enc_sz(s)
        qsz = dist.Normal(qsz_mu, self.softplus(qsz_logvar))
        sz = qsz.rsample()
        # deconst p
        p = self.spm(sz, xz)
        # construct mean for s
        sld = torch.matmul(p, xld)
        xld = xld * self.softplus(self.x_batch_rate[x_batch_idx]) + self.softplus(self.x_batch_add[x_batch_idx])
        sld = sld * self.softplus(self.s_batch_rate[x_batch_idx]) + self.softplus(self.s_batch_add[x_batch_idx])
        return(xz, qxz, xld, sz, qsz, p, sld)



# evndyn
import torch
import numpy as np

class VaeSmDataSet(torch.utils.data.Dataset):
    def __init__(self, x, xnorm_mat, transform=None, pre_transform=None):
        self.x = x
        self.xnorm_mat = xnorm_mat

    def __len__(self):
        return(self.x.shape[0])

    def __getitem__(self, idx):
        idx_x = self.x[idx]
        idx_xnorm_mat = self.xnorm_mat[idx]
        return(idx_x, idx_xnorm_mat)


class VaeSmDataSetMB(torch.utils.data.Dataset):
    def __init__(self, x, xnorm_mat, batch_idx, transform=None, pre_transform=None):
        self.x = x
        self.xnorm_mat = xnorm_mat
        self.batch_idx = batch_idx

    def __len__(self):
        return(self.x.shape[0])

    def __getitem__(self, idx):
        idx_x = self.x[idx]
        idx_xnorm_mat = self.xnorm_mat[idx]
        idx_batch_idx = self.batch_idx[idx]
        return(idx_x, idx_xnorm_mat, idx_batch_idx)


class ConcatDataset(torch.utils.data.Dataset):
    def __init__(self, *datasets):
        self.datasets = datasets

    def __getitem__(self, idx):
        data_list = [
            dataset[idx]
            for dataset in self.datasets]
        return(data_list)

    def __len__(self):
        min_length = min([
            len(dataset)
            for dataset in self.datasets])
        return(min_length)
    

class VaeSmDataManager():
    def __init__(self, x, test_ratio, batch_size, num_workers, validation_ratio=0.1):
        x = x.float()
        xnorm_mat = torch.mean(x, dim=1).view(-1, 1)
        self.x = x
        self.xnorm_mat = xnorm_mat
        total_num = x.size()[0]
        validation_num = int(total_num * validation_ratio)
        test_num = int(total_num * test_ratio)
        np.random.seed(42)
        idx = np.random.permutation(np.arange(total_num))
        validation_idx, test_idx, train_idx = idx[:validation_num], idx[validation_num:(validation_num +  test_num)], idx[(validation_num +  test_num):]
        self.validation_idx, self.test_idx, self.train_idx = validation_idx, test_idx, train_idx
        self.validation_x = x[validation_idx]
        self.validation_xnorm_mat = xnorm_mat[validation_idx]
        self.test_x = x[test_idx]
        self.test_xnorm_mat = xnorm_mat[test_idx]
        self.train_eds = VaeSmDataSet(x[train_idx], xnorm_mat[train_idx])
        self.train_loader = torch.utils.data.DataLoader(
            self.train_eds, batch_size=batch_size, shuffle=True, num_workers=num_workers, drop_last=True, pin_memory=True)

    def initialize_loader(self, batch_size, num_workers=2):
        self.train_loader = torch.utils.data.DataLoader(
            self.train_eds, batch_size=batch_size, shuffle=True, num_workers=num_workers, drop_last=True, pin_memory=True)


class VaeSmDataManagerMB():
    def __init__(self, x, batch_idx, test_ratio, batch_size, num_workers, validation_ratio=0.1):
        x = x.float()
        xnorm_mat = torch.mean(x, dim=1).view(-1, 1)
        self.x = x
        self.xnorm_mat = xnorm_mat
        self.batch_idx = batch_idx
        total_num = x.size()[0]
        validation_num = int(total_num * validation_ratio)
        test_num = int(total_num * test_ratio)
        np.random.seed(42)
        idx = np.random.permutation(np.arange(total_num))
        validation_idx, test_idx, train_idx = idx[:validation_num], idx[validation_num:(validation_num +  test_num)], idx[(validation_num +  test_num):]
        self.validation_idx, self.test_idx, self.train_idx = validation_idx, test_idx, train_idx
        self.validation_x = x[validation_idx]
        self.validation_xnorm_mat = xnorm_mat[validation_idx]
        self.validation_batch_idx = batch_idx[validation_idx]
        self.test_x = x[test_idx]
        self.test_xnorm_mat = xnorm_mat[test_idx]
        self.test_batch_idx = batch_idx[test_idx]
        self.train_eds = VaeSmDataSet(x[train_idx], xnorm_mat[train_idx], batch_idx[train_idx])
        self.train_loader = torch.utils.data.DataLoader(
            self.train_eds, batch_size=batch_size, shuffle=True, num_workers=num_workers, drop_last=True, pin_memory=True)

    def initialize_loader(self, batch_size, num_workers=2):
        self.train_loader = torch.utils.data.DataLoader(
            self.train_eds, batch_size=batch_size, shuffle=True, num_workers=num_workers, drop_last=True, pin_memory=True)


class VaeSmDataManagerDPP(VaeSmDataManager):
    def __init__(self, gpu, gpu_num, *args, **kwargs):
        super(VaeSmDataManagerDPP, self).__init__(*args, **kwargs)
        self.train_sampler = torch.utils.data.distributed.DistributedSampler(
            self.train_eds, num_replicas=gpu_num, rank=gpu)
        self.train_loader = torch.utils.data.DataLoader(
            self.train_eds, batch_size=args[2], shuffle=False, num_workers=0, drop_last=True, pin_memory=True, sampler=self.train_sampler)
        
    def initialize_loader(self, batch_size):
        self.train_loader = torch.utils.data.DataLoader(
            self.train_eds, batch_size=batch_size, shuffle=False, num_workers=0, drop_last=True, pin_memory=True, sampler=self.train_sampler)



import torch.distributions as dist


def calc_kld(qz):
    kld = -0.5 * (1 + qz.scale.pow(2).log() - qz.loc.pow(2) - qz.scale.pow(2))
    return(kld)


def calc_poisson_loss(ld, norm_mat, obs):
    p_z = dist.Poisson(ld * norm_mat)
    l = - p_z.log_prob(obs)
    return(l)
        
    
def calc_nb_loss(ld, norm_mat, theta, obs):
    ld = norm_mat * ld
    p =  ld / (ld + theta)
    p_z = dist.NegativeBinomial(theta, p)
    l = - p_z.log_prob(obs)
    return(l)
        



import torch
import numpy as np
from .exp import VaeSmExperiment
import scanpy as sc
import pandas as pd


def define_exp(
        x_fname, s_fname,
        model_params = {
            'x_dim': 100,
            'z_dim': 10,
            'enc_z_h_dim': 50, 'enc_d_h_dim': 50, 'dec_z_h_dim': 50,
            'num_enc_z_layers': 2, 'num_enc_d_layers': 2,
            'num_dec_z_layers': 2
        },
        lr=0.001, val_ratio=0.01, test_ratio=0.01, batch_ratio=0.05, num_workers=2, device='auto'):
    x = torch.tensor(np.loadtxt(x_fname))
    s = torch.tensor(np.loadtxt(s_fname))
    model_params['x_dim'] = x.size()[1]
    model_params['s_dim'] = s.size()[1]
    x_batch_size = int(x.size()[0] * batch_ratio)
    s_batch_size = int(s.size()[0] * batch_ratio)
    vaesm_exp = VaeSmExperiment(model_params, lr, x, s, test_ratio, 100, 100, num_workers, validation_ratio=val_ratio, device=device)
    return(vaesm_exp)


# defining useful function
def select_slide(adata, s, s_col='sample', s_sub=None):
    if s_sub == None:
        slide = adata[adata.obs[s_col].isin([s]), :]
    else:
        slide = adata[adata.obs[s_col].isin([s_sub]), :]
    s_keys = list(slide.uns['spatial'].keys())
    s_spatial = s_keys[0]
    slide.uns['spatial'] = {s_spatial: slide.uns['spatial'][s]}
    return slide


def conduct_umap(adata, key):
    sc.pp.neighbors(adata, use_rep=key, n_neighbors=30)
    sc.tl.umap(adata)
    return(adata)

def convert2array(mat):
    if isinstance(mat, np.ndarray):
        return(mat)
    else:
        return(mat.toarray())


def plot_mapped_sc(sc_adata, mapping, ax):
    embed = sc_adata.obsm['X_umap']
    ax.scatter(embed[:, 0], embed[:, 1], c='gray', s=5)
    ax.scatter(embed[:, 0], embed[:, 1], c=mapping, s=30 * mapping / np.max(mapping))

def calculate_roc_df(pred, target):
    stats_df = pd.DataFrame({'pred': pred, 'target': target})
    stats_df['pos_target'] = (stats_df['target'] > 0).astype(int)
    stats_df['neg_target'] = (stats_df['target'] == 0).astype(int)
    stats_df['pos_target'] = stats_df['pos_target'] / stats_df['pos_target'].sum()
    stats_df['neg_target'] = stats_df['neg_target'] / stats_df['neg_target'].sum()
    stats_df = stats_df.groupby('pred', as_index=False).sum()
    stats_df = stats_df.sort_values('pred', ascending=False)
    stats_df['tpr'] = np.cumsum(stats_df['pos_target'])
    stats_df['fpr'] = np.cumsum(stats_df['neg_target'])
    return(stats_df)


def calculate_auc(pred, target):
    stats_df = pd.DataFrame({'pred': pred, 'target': target})
    stats_df['norm_target'] = stats_df['target'].div(stats_df['target'].sum())
    stats_df['ompr'] = stats_df['pred'].rank(method='average').div(stats_df.shape[0])
    auc = (stats_df['norm_target'] * stats_df['ompr']).sum()
    return(auc)
    
def calculate_recall(pred, target, q):
    thresh = np.quantile(pred, 1 - q)
    target = target / np.sum(target)
    recall = np.sum(target[pred > thresh])
    return(recall)

def process_each_ensembl(symbol, info_df):
    val = info_df.loc[symbol]['ensembl']
    if type(val) == dict:
        return([val['gene']])
    elif type(val) == list:
        ensembls = [d['gene'] for d in val]
        return(ensembls)
    else:
        return([])
        


def cut_unmapped(adata, q):
    contrib_vec = np.sort(adata.obsm['map2sp'].sum(axis=1))
    cum_contrib_vec = np.cumsum(contrib_vec)
    cum_contrib_vec = cum_contrib_vec / np.max(cum_contrib_vec)
    map_thresh = np.max(contrib_vec[cum_contrib_vec < q])
    adata = adata[adata.obsm['map2sp'].sum(axis=1) > map_thresh]
    return(adata)
    

def make_celltype_coloc(sc_adata, celltypes, celltype_label, thresh=2.0):
    mapped_cells = cut_unmapped(sc_adata, 0.05).obs_names
    map_vec = sc_adata.obs_names.isin(mapped_cells).astype(int)
    p_mat = sc_adata.obsm['map2sp'] / np.sum(sc_adata.obsm['map2sp'], axis=1).reshape((-1, 1))
    p_mat = p_mat * map_vec.reshape((-1, 1))
    coloc_mat = p_mat @ p_mat.transpose()
    coloc_mat = coloc_mat * p_mat.shape[1]
    bcoloc_mat = (coloc_mat > thresh).astype(int)
    celltype_coloc_props = np.array(
        [[
            np.sum(coloc_mat[sc_adata.obs[celltype_label] == celltype1][:, sc_adata.obs[celltype_label] == celltype2]) /
            (np.sum(sc_adata.obs[celltype_label] == celltype1) * np.sum(sc_adata.obs[celltype_label] == celltype2)) 
            for celltype2 in celltypes]
         for celltype1 in celltypes])
    return(celltype_coloc_props)

def make_df_col_category(df, col, categories):
    df[col] = pd.Categorical(df[col], categories=categories)
    return(df)

def categolize_method(methods):
    methods_cat = pd.Categorical(methods, categories=['scoloc', 'cell2loc', 'tangram'])
    return(methods_cat)


def make_edge_df(sc_adata, large_celltype_label, sub_sample=True, tot_size=5000, exclude_reverse=True, edge_thresh=1):
    if sub_sample:
        tot_size = 5000
        sub_sc_adata = sc_adata[np.random.choice(sc_adata.obs_names, tot_size, replace=False)]
    else:
        sub_sc_adata = sc_adata
    p_mat = sub_sc_adata.obsm['map2sp'] / np.sum(sub_sc_adata.obsm['map2sp'], axis=1).reshape((-1, 1))
    coloc_mat = p_mat @ p_mat.transpose()
    coloc_mat = np.log2(coloc_mat) + np.log2(p_mat.shape[1])
    thresh = edge_thresh
    ## thresh = np.quantile(coloc_mat, 0.8)
    high_coloc_index = np.argwhere(coloc_mat > thresh)
    if exclude_reverse:
        high_coloc_index = high_coloc_index[high_coloc_index[:, 0] < high_coloc_index[:, 1]]
    ocell1_types = sub_sc_adata.obs[large_celltype_label].iloc[high_coloc_index[:, 0]].values
    ocell2_types = sub_sc_adata.obs[large_celltype_label].iloc[high_coloc_index[:, 1]].values
    high_coloc_index = high_coloc_index[ocell1_types != ocell2_types]
    cell1_types = ocell1_types[ocell1_types != ocell2_types]
    cell2_types = ocell2_types[ocell1_types != ocell2_types]
    edge_idx = np.arange(cell1_types.shape[0])
    orig_edge_df = pd.DataFrame({'edge': edge_idx, 'cell1': sub_sc_adata.obs_names[high_coloc_index[:, 0]], 'cell2': sub_sc_adata.obs_names[high_coloc_index[:, 1]], 'cell1_type': cell1_types, 'cell2_type': cell2_types}, index=edge_idx)
    return(orig_edge_df)


def calc_signature_score(adata, orig_genes,label,  min_count=100):
    genes = orig_genes[np.isin(orig_genes, adata.var_names)]
    tot_vec = np.array(adata[:, genes].layers['count'].sum(axis=0))
    genes = genes[tot_vec > min_count]
    sc.tl.score_genes(adata, genes, score_name=label)
    return(adata)


def trancate_ext_val(vec, q=0.01):
    high_val = np.quantile(vec, 1 - q)
    vec[vec > high_val] = high_val
    return(vec)

def make_count_vec(X, axis):
    count_vec = np.array(X.sum(axis=axis)).reshape(-1)
    return(count_vec)
